use crate::*;
use serde::{Deserialize, Serialize};
use std::collections::HashMap;
use std::fmt;
use std::io::{stdout, Write};
use std::path::PathBuf;
use std::process::Command;

#[derive(Debug, Serialize, Deserialize, PartialEq)]
enum UpdateSteps {
    Fetch,
    Compile,
    Install,
}

#[derive(Debug, Serialize, Deserialize)]
pub struct Updater {
    systems: Vec<System>,
}

#[derive(Debug, Serialize, Deserialize)]
pub struct System {
    //name: String,
    fetch: Option<Cmd>,
    compile: Option<Cmd>,
    install: Cmd,
    // deps or rDeps : Tree
    // exclusive_with : List
}

#[derive(Debug, Serialize, Deserialize, Clone)]
struct Cmd {
    exe: String,
    params: Vec<String>,
    yes: Option<String>,
    current_dir: Option<PathBuf>,
    env: HashMap<String, String>,
}

#[derive(Debug, Serialize, Deserialize)]
struct ActualCmd {
    exe: String,
    params: Vec<String>,
    current_dir: Option<PathBuf>,
    env: HashMap<String, String>,
}

// QUESTION: Would String be less awkward?
impl From<&str> for UpdateSteps {
    fn from(value: &str) -> Self {
        match &*value.to_lowercase() {
            "fetch" => UpdateSteps::Fetch,
            "compile" => UpdateSteps::Compile,
            "install" => UpdateSteps::Install,

            _ => panic!("Step {} not recognized", value),
        }
    }
}

// QUESTION: Would String be less awkward?
impl From<&String> for UpdateSteps {
    fn from(value: &String) -> Self {
        match &*value.to_lowercase() {
            "fetch" => UpdateSteps::Fetch,
            "compile" => UpdateSteps::Compile,
            "install" => UpdateSteps::Install,

            _ => panic!("Step {} not recognized", value),
        }
    }
}

impl Updater {
    fn new() -> Updater {
        let mut up = Updater { systems: vec![] };
        up.systems.push(System {
            fetch: None,
            compile: None,
            install: Cmd::new(),
        });
        up
    }

    /// To create a sample config from code
    #[doc(hidden)]
    fn write_config(&self, opt: &Opt) {
        use std::fs::OpenOptions;

        let mut f = OpenOptions::new()
            .write(true)
            .create(true)
            .truncate(true)
            .open(&opt.config_file)
            .unwrap();

        f.write_all(serde_yaml::to_string(&self).unwrap().as_bytes())
            .unwrap();
    }

    pub fn from_config(opt: &Opt) -> Result<Updater> {
        // let u = Updater::new();
        // u.write_config(opt);

        let file = std::fs::read_to_string(&opt.config_file).unwrap();
        Ok(serde_yaml::from_str(&file).unwrap())
    }

    pub fn update_all(&self, opt: &Opt) -> Result<()> {
        let errors: Vec<_> = self
            .systems
            .iter()
            .filter_map(|sys| {
                if let Err(err) = self.update(sys, opt) {
                    eprintln!("Error catched {}", err);
                    Some(err)
                } else {
                    None
                }
            })
            .collect();

        // TODO:
        if errors.is_empty() {
            Ok(())
        } else {
            Err(Error::new(
                errors,
                ErrorKind::Fetch, // TODO: Why should I choose here?
            ))
        }
    }

    fn update(&self, sys: &System, opt: &Opt) -> Result<()> {
        let steps = match &opt.steps {
            Some(v) => v.iter().map(|u| u.into()).collect(),
            None => vec![
                UpdateSteps::Fetch,
                UpdateSteps::Compile,
                UpdateSteps::Install,
            ],
        };

        if steps.contains(&UpdateSteps::Fetch) {
            sys.fetch(opt)?;
        }
        if steps.contains(&UpdateSteps::Compile) {
            sys.compile(opt)?;
        }
        if steps.contains(&UpdateSteps::Install) {
            sys.install(opt)?;
        }

        Ok(())
    }
}

impl System {
    pub fn fetch(&self, opt: &Opt) -> Result<()> {
        if let Some(fetch) = &self.fetch {
            fetch.clone().prepare(opt).execute(opt)?;
        }
        Ok(())
    }

    pub fn compile(&self, opt: &Opt) -> Result<()> {
        if let Some(compile) = &self.compile {
            compile.clone().prepare(opt).execute(opt)?;
        }
        Ok(())
    }

    pub fn install(&self, opt: &Opt) -> Result<()> {
        self.install.clone().prepare(opt).execute(opt)?;
        Ok(())
    }
}

impl Cmd {
    fn new() -> Cmd {
        Cmd {
            exe: "".into(),
            params: vec![],
            yes: None,
            current_dir: None,
            env: HashMap::new(),
        }
    }

    fn prepare(self, opt: &Opt) -> ActualCmd {
        let mut params = self.params;
        if opt.yes {
            if let Some(y) = &self.yes {
                params.push(y.to_owned());
            }
        };

        let mut env = self.env;
        if !env.contains_key("PATH") {
            env.insert("PATH".to_owned(), std::env!("PATH").to_owned());
        }

        ActualCmd {
            exe: self.exe,
            params,
            current_dir: self.current_dir,
            env,
        }
    }
}

impl ActualCmd {
    fn execute(&self, opt: &Opt) -> Result<bool> {
        let mut cmd = Command::new(&self.exe);

        cmd.args(&self.params)
            .env_clear()
            .envs(&self.env)
            // .stdout(cfg)
            // .stderr(cfg)
            ;

        if let Some(cdir) = &self.current_dir {
            cmd.current_dir(std::fs::canonicalize(cdir).unwrap());
        }

        let confirm = if opt.yes { "\n" } else { "[Y/n]? " };

        print!("Will execute {}{}", self, confirm);
        stdout().flush().unwrap();

        let mut yes_no = String::new();

        if !opt.yes {
            std::io::stdin()
                .read_line(&mut yes_no)
                .expect("Unable to read user’s input");
        }

        if yes_no.to_lowercase().starts_with('n') {
            eprintln!(
                "yes_no? {} {}",
                yes_no,
                yes_no.to_lowercase().starts_with('n')
            );
            return Ok(false);
        }

        Ok(cmd.status().unwrap().success())
    }
}

impl fmt::Display for ActualCmd {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        let command = if !self.params.is_empty() {
            format!("{} {}", &self.exe, &self.params.join(" "))
        } else {
            self.exe.clone()
        };
        write!(f, "`{}`", command)?;

        if let Some(cdir) = &self.current_dir {
            write!(f, " in {:?}", cdir)?;
        }

        if !self.env.is_empty() {
            writeln!(f, " with the following environment variable:")?;
            writeln!(f, "{:#?}", self.env)
        } else {
            write!(f, " without any environment variable. ")
        }
    }
}
